"""
exceptions.py – Tabletx exceptions
"""

class TabletException(Exception):
    pass

class BadConfigData(TabletException):
    pass

class NonSystemInitialLayer(TabletException):
    pass

class TabletBoundsExceeded(TabletException):
    pass



""" scene_view.py - create a vector drawing view in QT """

import logging
from PyQt6.QtWidgets import QGraphicsView, QVBoxLayout, QGraphicsScene, QWidget
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor, QPainter, QPageSize, QPageLayout
from PyQt6.QtPrintSupport import QPrinter
from pypdf import PdfWriter, PdfReader

_logger = logging.getLogger(__name__)


class SceneView(QGraphicsView):
    def __init__(self, size, background):
        super().__init__()

        # Create a QGraphicsScene
        self.scene = QGraphicsScene()
        self.setSceneRect(0, 0, size.width, size.height)
        pad = 10
        self.setFixedSize(size.width + pad, size.height + pad)
        if background:
            _logger.info(f"Setting scene background to RGB: {background}")
            self.scene.setBackgroundBrush(QColor(*background))

        # Set the scene to the view
        self.setScene(self.scene)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        # self.setRenderHint(QPainter.RenderHint.Antialiasing)
        # self.setRenderHint(QPainter.RenderHint.TextAntialiasing)

    def save_as_pdf(self, file_path):
        # Create a QPrinter object and set the output format to PDF
        printer = QPrinter(QPrinter.PrinterMode.PrinterResolution)
        printer.setOutputFormat(QPrinter.OutputFormat.PdfFormat)
        # printer.setPageSize(QPageSize(self.size))
        ps = QPageSize(QPageSize.PageSizeId.AnsiC)
        printer.setPageSize(ps)
        printer.setPageOrientation(QPageLayout.Orientation.Landscape)
        # printer.setPageMargins(QMarginsF(0, 0, 0, 0))
        printer.setFullPage(True)
        printer.setOutputFileName(str(file_path))

        # Create a QPainter to render the view content onto the printer
        painter = QPainter(printer)

        # Render the entire scene to the printer (the PDF)
        self.render(painter)

        # Finish the painting process
        painter.end()

        # self.crop_pdf(file_path)

    def crop_pdf(self, file_path):
        reader = PdfReader(file_path)
        writer = PdfWriter()
        crop = 10

        sheet = reader.pages[0]
        crop_box = sheet.cropbox
        crop_box.lower_left = (crop_box.lower_left[0], crop_box.lower_left[1] + crop)
        sheet.cropbox.lower_left = crop_box.lower_left
        writer.add_page(sheet)
        with open("cropped.pdf", "wb") as fp:
            writer.write(fp)


class MainWindow(QWidget):
    def __init__(self, title, size, background):
        super().__init__()

        # Set up the layout and graphics view
        layout = QVBoxLayout()
        self.graphics_view = SceneView(size, background)
        layout.addWidget(self.graphics_view)

        self.setLayout(layout)
        self.setWindowTitle(title)
        self.setGeometry(500, 100, size.width, size.height)  # Position and size of the main window on the screen



"""
styledb.py - Loads styles common to all Presentations
"""
import logging
from pathlib import Path
import yaml
from collections import namedtuple
from tabletqt.exceptions import BadConfigData

_logger = logging.getLogger(__name__)

Color_Canvas = namedtuple('Color_Canvas', 'r g b canvas')
Float_RGB = namedtuple('Float_RGB', 'r g b')
Line_Style = namedtuple('Line_Style', 'pattern width color')
Text_Style = namedtuple('Text_Style', 'typeface size slant weight color spacing')
Dash_Pattern = namedtuple('Dash_Pattern', 'solid blank')

config_dir = Path(__file__).parent / "config"
# nt - named tuple is defined
# pre/post - Whether or not the data must be pre or post processed
PP = namedtuple('PP', 'nt pre post')  # Postprocess config file data

# yaml file name : load data using this tuple
config_type = {
    'colors': PP(Color_Canvas, pre=True, post=False),
    'line_styles': PP(nt=Line_Style, pre=False, post=False ),
    'dash_patterns': PP(nt=Dash_Pattern, pre=False, post=False),
    'typefaces': PP(nt=None, pre=False, post=False),
    'text_styles': PP(Text_Style, pre=False, post=True),
    'color_usages': PP(nt=None, pre=None, post=True)
}


def load_yaml_to_namedtuple(file_path, namedtuple_type):
    with open(file_path, 'r') as file:
        raw_data = yaml.safe_load(file)
    if not isinstance(raw_data, dict):
        raise BadConfigData(f"Expected dict when loading:\n    {file_path}")
    return {k: namedtuple_type(**v) for k, v in raw_data.items()}


class StyleDB:
    """
    Singleton class interface to the Presentation and Styles in the Flatland database. Created with an initial
    Presentation and loads all presentation/style data for that Presentation for easy access by
    the Tablet.
    """
    styles = {}  # { 'text_style':
    rgbF = {}  # rgb color float representation
    typeface = None
    dash_pattern = None
    line_style = None
    text_style = None
    color_usage = None

    @classmethod
    def load_config_files(cls):
        """
        Processes the config_type dictionary, loading each yaml config file into either
        a named tuple or a simple key value dictionary if no named tuple is provided
        and then sets the corresponding StyleDB class attribute to that value
        """
        _logger.info(f"StyleDB loading tabletqt configuration\n---")
        for fname, pp in config_type.items():
            config_file_path = config_dir / (fname+".yaml")
            _logger.info(f"loading: {config_file_path}")
            if pp.nt:
                # A named tuple is defined for this file, so we use it to load the data
                attr_val = load_yaml_to_namedtuple(config_file_path, pp.nt)
            else:
                # No named tuple, so we just read file into a dictionary
                with open(config_file_path, 'r') as file:
                    attr_val = yaml.safe_load(file)
            if pp.pre:
                # Retrieve the relevant preprocessing method name using the filename suffix
                method_name = 'preprocess_'+fname
                method = getattr(cls, method_name, None)
                # Invoke it on the loaded yaml data
                method(attr_val)
            attr_name = fname[:-1]  # drop the plural 's' from the file name to get the attribute name
            # Assign the loaded and possibly preprocessed yaml data to the relevant class attribute
            setattr(cls, attr_name, attr_val)
            if pp.post:
                method_name = 'postprocess_'+fname  # Keep the plural
                method = getattr(cls, method_name, None)
                method()

        _logger.info(f"---\n")

    @classmethod
    def postprocess_text_styles(cls):
        """
        Verify all typefaces are defined
        """
        undefined_typefaces = [t.typeface for t in cls.text_style.values() if t.typeface not in cls.typeface]
        if undefined_typefaces:
            _logger.error(f"Undefined typefaces: {undefined_typefaces} encountered in"
                          f"text styles config file:\n    {config_dir / 'text_styles.yaml'}")
            raise BadConfigData


    @classmethod
    def preprocess_colors(cls, raw_data):
        """
        # Validate color float values for QT6
        """
        for name, rgb in raw_data.items():
            for n in [rgb.r, rgb.g, rgb.b]:
                if not 0 <= n <= 255:
                    _logger.error(f"Bad color value [{n}] for: {name} in "
                                  f"config file:\n    {config_dir / 'colors.yaml'}")
                    raise BadConfigData
            StyleDB.rgbF[name] = Float_RGB(r=rgb.r, g=rgb.g, b=rgb.b)

    @classmethod
    def postprocess_color_usages(cls):
        """
        Validate color names
        """
        undefined_colors = [c for c in cls.color_usage.values() if c not in cls.rgbF]
        if undefined_colors:
            _logger.error(f"Undefined colors: {undefined_colors} encountered in"
                          f"color usages config file:\n    {config_dir / 'color_usages.yaml'}")
            raise BadConfigData

    @classmethod
    def report_colors(cls):
        """
        Prints out a list of the available colors for the user
        """
        cls.load_config_files()
        print("Canvas colors:")
        print("---")
        for c in cls.rgbF:
            print(c)
        print("===")



"""
presentation.py – Presentation class in View domain
"""
import logging
from collections import namedtuple
import yaml
from pathlib import Path


CornerSpec = namedtuple('Corner_Spec', 'radius top bottom')
config_dir = Path(__file__).parent / "config"


class Presentation:
    """
   A set of compatible visual styles including fonts, colors, border widths and so forth as appropriate to a
   given View Type form a selectable Presentation. For example, an `Executable UML State Machine Diagram`
   might be drawn using certain fonts for state names and possibly different colors for transient and
   non-transient states. Alternatively, only black and white might be used with purple for a certain kind of
   connector in a diagnostic Presentation.
   """

    def __init__(self, name: str, drawing_type: str):
        """
       Constructor
       """
        self.logger = logging.getLogger(__name__)
        self.Name = name
        self.Drawing_type = drawing_type
        self.Text_presentation = {}
        self.Underlays = set()  # Set of text presentations that require an underlay
        self.Shape_presentation = {}
        self.Closed_shape_fill = {}
        self.Corner_spec = {}

        # Load Asset Presentations for all Assets in this Presentation
        self.logger.info(f"using presentation: [{self.Name}]")
        self.load_drawing_type()

    def load_drawing_type(self):
        """
        Load shape and text assets for my drawing type and presentation
        """
        file_path = config_dir / "drawing_types.yaml"
        self.logger.info(f"loading assets in: [{file_path}]")
        with open(file_path, 'r') as file:
            raw_data = yaml.safe_load(file)
        my_data = raw_data[self.Drawing_type][self.Name]
        # Load text presentations
        for asset_name,v in my_data['text'].items():
            self.Text_presentation[asset_name] = v['Text style']
            if v['Underlay']:
                self.Underlays.add(asset_name)
        # Load shape presentations
        for asset_name, v in my_data['shape'].items():
            self.Shape_presentation[asset_name] = v['Line style']
            fill = v.get('Fill')
            if fill:
                self.Closed_shape_fill[asset_name] = fill
            corner_spec = v.get('Corner spec')
            if corner_spec:
                self.Corner_spec[asset_name] = CornerSpec(radius=corner_spec['Radius'], top=corner_spec['Top'],
                                                          bottom=corner_spec['Bottom'])
